with Ada.Calendar,
     Ada.Characters.Handling,
     Ada.Text_Io,
     Tty_Out;

package body Statistics is
  use type Claw.Dword;
  use type Ada.Calendar.Time;
  Progress_Interval: constant Duration := 5.0; -- update screen period
  Start_Time: constant Ada.Calendar.Time := Ada.Calendar.Clock;
  Next_Update_Time: Ada.Calendar.Time := Start_Time + Progress_Interval;
  type Class_Type is (Html_Text,  -- HTML file
                      Graphic,    -- GIF, JPG files
                      Text,       -- Text (TXT) file
                      Pdf,        -- PDF file
                      Postscript, -- PS file
                      Ms_Word,    -- DOC file
                      Archived,   -- ZIP, GZ, tar, etc.
                      Unclassified); -- Other files.

  function My_Class(Lower_Url: String) return Class_Type is
  -- Return the class of the URL, if possible. The URL is in lower case.
  -- Note that we do a dumb end-of-URL check here, which avoids the
  -- compilations and just dumps everything odd into unclassified.
  begin
    if Lower_Url'Length < 6 then
      return Unclassified;
    elsif Lower_Url(Lower_Url'Last - 2 .. Lower_Url'Last) = ".gz" then
      return Archived;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".zip" then
      return Archived;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".tar" then
      return Archived;
    elsif Lower_Url(Lower_Url'Last - 1 .. Lower_Url'Last) = ".z" then
      -- Includes .tar.Z
      return Archived;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".taz" then
      -- Really .tar.Z for MS-DOS.
      return Archived;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".htm" then
      return Html_Text;
        -- Note: We could look at the returned type for this one.
    elsif Lower_Url(Lower_Url'Last - 4 .. Lower_Url'Last) = ".html" then
      return Html_Text;
    elsif Lower_Url(Lower_Url'Last - 5 .. Lower_Url'Last) = ".shtml" then
      return Html_Text;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".pdf" then
      return Pdf;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".txt" then
      return Text;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".doc" then
      return Ms_Word;
    elsif Lower_Url(Lower_Url'Last - 2 .. Lower_Url'Last) = ".ps" then
      return Postscript;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".bmp" then
      return Graphic;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".gif" then
      return Graphic;
    elsif Lower_Url(Lower_Url'Last - 3 .. Lower_Url'Last) = ".jpg" then
      return Graphic;
    elsif Lower_Url(Lower_Url'Last - 4 .. Lower_Url'Last) = ".jpeg" then
      return Graphic;
    else
      return Unclassified;
    end if;
  end My_Class;

  -- Statistics variables:
  Total_Count: Claw.Dword := 0;
  Foreign_Count: Claw.Dword := 0;
  Foreign_Failed_Count: Claw.Dword := 0;
  Local_Failed_Count: Claw.Dword := 0;
  Foreign_Size: Claw.Dword := 0;
  Local_Size: Claw.Dword := 0;
  Local_K_Size: Claw.Dword := 0;
  Class_Count: array (Class_Type) of Claw.Dword := (others => 0);
  Class_Size: array (Class_Type) of Claw.Dword := (others => 0);
  Class_K_Size: array (Class_Type) of Claw.Dword := (others => 0);
  Foreign_Scanned_Count: Claw.Dword := 0;
  Local_Scanned_Count: Claw.Dword := 0;
  Local_Scanned_Size: Claw.Dword := 0;
  Links_Scanned: Claw.Dword := 0;
  Old_Domain_Links: Claw.Dword := 0;
  Adaic_Links: Claw.Dword := 0;
  Bad_Links: Claw.Dword := 0;
  Retry_Success_Count: array (Boolean) of Claw.Dword := (others => 0);

  procedure Increment_Links_Scanned is
  begin
    Links_Scanned := Links_Scanned + 1;
  end Increment_Links_Scanned;

  procedure Increment_Old_Domain_Links is
  begin
    Old_Domain_Links := Old_Domain_Links + 1;
  end Increment_Old_Domain_Links;

  procedure Increment_Adaic_Links is
  begin
    Adaic_Links := Adaic_Links + 1;
  end Increment_Adaic_Links;

  procedure Increment_Bad_Links is
  begin
    Bad_Links := Bad_Links + 1;
  end Increment_Bad_Links;

  procedure Finished_One(Size    : in   Claw.Dword; -- finished with some url
                         Foreign : in   Boolean;
                         Failed  : in   Boolean;
                         Scanned : in   Boolean;
                         Url_String: in String;
                         Second_Try: in Boolean;
                         Url_Count: in  Claw.Dword) is
    Class   : constant Class_Type
      := My_Class(Ada.Characters.Handling.To_Lower(Url_String));
  begin
    -- Save statistics:
    if Second_Try then
      Retry_Success_Count(not Failed) := Retry_Success_Count(not Failed) + 1;
    end if;
    if Failed and Second_Try then
      null; -- Do not change statistics.
    elsif Second_Try then
      -- The second try succeeded, decrease the number of
      -- failures, but don't increase the raw count.
      --Total_Count unchanged.
      if Foreign then
        --Foreign_Count unchanged.
        Foreign_Failed_Count := Foreign_Failed_Count - 1;
        if Scanned then
          Foreign_Scanned_Count := Foreign_Scanned_Count + 1;
          -- Shouldn't be non-zero!
        end if;
        Foreign_Size := Foreign_Size + Size;
      else
        Local_Size := Local_Size + Size;
        Local_K_Size := Local_K_Size + (Size + 1023) / 1024;
        Local_Failed_Count := Local_Failed_Count - 1;
        Class_Count(Class) := Class_Count(Class) + 1;
        Class_Size(Class) := Class_Size(Class) + Size;
        Class_K_Size(Class) := Class_K_Size(Class) + (Size + 1023) / 1024;
        if Scanned then
          Local_Scanned_Count := Local_Scanned_Count + 1;
          Local_Scanned_Size := Local_Scanned_Size + Size;
          -- Note that this isn't really useful, as it doesn't
          -- include the size of items scanned which don't
          -- have fixed sizes (e.g. directories).
        end if;
      end if;
    else
      -- First_Try.
      Total_Count := Total_Count + 1;
      if Foreign then
        Foreign_Count := Foreign_Count + 1;
        if Failed then
          Foreign_Failed_Count := Foreign_Failed_Count + 1;
        end if;
        if Scanned then
          Foreign_Scanned_Count := Foreign_Scanned_Count + 1;
          -- Shouldn't be non-zero!
        end if;
        Foreign_Size := Foreign_Size + Size;
      else
        Local_Size := Local_Size + Size;
        Local_K_Size := Local_K_Size + (Size + 1023) / 1024;
        if Failed then
          Local_Failed_Count := Local_Failed_Count + 1;
        else
          Class_Count(Class) := Class_Count(Class) + 1;
          Class_Size(Class) := Class_Size(Class) + Size;
          Class_K_Size(Class) := Class_K_Size(Class) + (Size + 1023) / 1024;
        end if;
        if Scanned then
          Local_Scanned_Count := Local_Scanned_Count + 1;
          Local_Scanned_Size := Local_Scanned_Size + Size;
          -- Note that this isn't really useful, as it doesn't
          -- include the size of items scanned which don't
          -- have fixed sizes (e.g. directories).
        end if;
      end if;
    end if;
    if Ada.Calendar.Clock >= Next_Update_Time then
      -- Occasional progress reports to Standard Output:
      Display(Url_Count);
      Next_Update_Time := Next_Update_Time + Progress_Interval;
    end if;
  end Finished_One;

  function Fancy_Image(I       : Claw.Dword;
                       Width   : Natural := 0) return String is
  -- Return number with commas in it; the length will be Width or
  -- more characters.
    Number  : constant String := Claw.Dword'Image(I);
    -- Shortest possible Number is two characters.
    Comma_Number: String(1 .. Natural'Max(Number'Length + (Number'Length-2)/3,
                                          Width))
      := (others => ' ');
    J : Natural := Comma_Number'Last;
  begin
    Comma_Number(1) := Number(1);
      -- Copy Sign (should always be a space, as Claw.DWord is unsigned).
    for I in reverse 2 .. Number'Last loop
      if I /= Number'Last and (Number'Last - I) mod 3 = 0 then
        Comma_Number(J) := ',';
        J := J - 1;
      end if;
      Comma_Number(J) := Number(I);
      J := J - 1;
    end loop;
    return Comma_Number;
  end Fancy_Image;

  Screen  : Tty_Out.Tty_Type;

  procedure Set_Name(Name : in String) is
  -- set the name of the operator's monitoring window
  begin
    Tty_Out.Put_Line(Screen, "Scan " & Name); -- ensure window opened
    Tty_Out.Set_Name(Screen, Name); -- raises Not_Valid_Error if closed
    Tty_Out.Move(Screen, Size => (Width => 480, Height => 248));
  end Set_Name;

  function Kill_Request return Boolean is
  -- True iff op wants to stop run
  begin
    return Tty_Out.Is_Closed(Screen);
  end Kill_Request;

  procedure Display(Url_Count: in    Claw.Dword) is
  begin
    Tty_Out.New_Page(Screen);
    Tty_Out.Put_Line(Screen, "%% So far: Stored" & Fancy_Image(Url_Count)
      & " URLs; processed" & Fancy_Image(Total_Count) & " URLs");
    Tty_Out.Put_Line(Screen, "%% So far: Local   "
      & Fancy_Image(Total_Count - Foreign_Count, 7)
      & " URLs processed, of which" & Fancy_Image(Local_Failed_Count, 4)
      & " failed");
    Tty_Out.Put_Line(Screen, "%% So far: Foreign "
      & Fancy_Image(Foreign_Count, 7) & " URLs processed, of which"
      & Fancy_Image(Foreign_Failed_Count, 4) & " failed");
    Tty_Out.Put_Line(Screen, "%% So far: Local Size " & Fancy_Image(Local_Size)
      & " bytes; or" & Fancy_Image(Local_K_Size) & " KBytes");
    Tty_Out.Put_Line(Screen, "  Scanned:" & Fancy_Image(Local_Scanned_Count)
      & " URLs;" & Fancy_Image(Links_Scanned)
      & " Links;" & Fancy_Image(Local_Scanned_Size) & " bytes");
    Tty_Out.Put_Line(Screen, "  Bad: " & Fancy_Image(Bad_Links) & " Links;"
      & " sw-eng domain converted:" & Fancy_Image(Old_Domain_Links)
      & " Links; AdaIC prefix: " & Fancy_Image(Adaic_Links) & " Links");
    for I in Class_Type loop
      case I is
        when Html_Text =>
          Tty_Out.Put(Screen, "HTML files              ");
        when Archived =>
          Tty_Out.Put(Screen, "Archived (ZIP, tar, gz) ");
        when Pdf =>
          Tty_Out.Put(Screen, "PDF files               ");
        when Ms_Word =>
          Tty_Out.Put(Screen, "Microsoft Word (.doc)   ");
        when Text =>
          Tty_Out.Put(Screen, "Text files              ");
        when Postscript =>
          Tty_Out.Put(Screen, "Postscript files        ");
        when Graphic =>
          Tty_Out.Put(Screen, "Graphic files           ");
        when Unclassified =>
          Tty_Out.Put(Screen, "Unclassified links      ");
      end case;
      Tty_Out.Put_Line(Screen, "Count:" & Fancy_Image(Class_Count(I), 7)
        & " Size:" & Fancy_Image(Class_Size(I), 12) & " or"
        & Fancy_Image(Class_K_Size(I), 9) & "K");
    end loop;
  end Display;

  procedure Print(Result_File: in  Ada.Text_Io.File_Type) is
    use type Ada.Calendar.Time;
  begin
    Ada.Text_Io.Put_Line(Result_File, "%% Total URLs processed: "
      & Fancy_Image(Total_Count, 12));
    Ada.Text_Io.Put_Line(Result_File, "%% Foreign:");
    Ada.Text_Io.Put_Line(Result_File, "%%   URLs processed:     "
      & Fancy_Image(Foreign_Count, 12));
    Ada.Text_Io.Put_Line(Result_File, "%%   URLs failed:        "
      & Fancy_Image(Foreign_Failed_Count, 12));
    if Foreign_Scanned_Count /= 0 then
      Ada.Text_Io.Put_Line(Result_File, "%%   URLs scanned:       "
        & Fancy_Image(Foreign_Scanned_Count, 12));
    end if;
    if Foreign_Size /= 0 then
      Ada.Text_Io.Put_Line(Result_File, "%%   Size:               "
        & Fancy_Image(Foreign_Size, 12) & " bytes");
    end if;
    Ada.Text_Io.Put_Line(Result_File, "%% Scanned site:");
    Ada.Text_Io.Put_Line(Result_File, "%%   URLs processed:     "
      & Fancy_Image(Total_Count - Foreign_Count, 12));
    Ada.Text_Io.Put_Line(Result_File, "%%   URLs failed:        "
      & Fancy_Image(Local_Failed_Count, 12));
    Ada.Text_Io.Put_Line(Result_File, "%%   Total size:         "
      & Fancy_Image(Local_Size, 12) & " bytes");
    Ada.Text_Io.Put_Line(Result_File, "%%   Size rounded up:    "
      & Fancy_Image(Local_K_Size, 12) & "K");
    Ada.Text_Io.Put_Line(Result_File, "%%   "
      & Claw.Dword'Image(Retry_Success_Count(True))
      & "/"
      & (Claw.Dword'Image(Retry_Success_Count(True)+Retry_Success_Count(False)))
      & " successful retries");
    Ada.Text_Io.Put_Line(Result_File, "%%   Elapsed time:"
      & Duration'Image(Ada.Calendar.Clock - Start_Time) & " seconds");
    Ada.Text_Io.Put_Line(Result_File, "%%   Scanned:"
      & Fancy_Image(Local_Scanned_Count) & " URLs;"
      & Fancy_Image(Links_Scanned) & " Links;"
      & Fancy_Image(Local_Scanned_Size) & " bytes");
    Ada.Text_Io.Put_Line(Result_File, "%%   Bad: " & Fancy_Image(Bad_Links)
      & " Links;" & " sw-eng domain converted:" & Fancy_Image(Old_Domain_Links)
      & " Links; AdaIC prefix: " & Fancy_Image(Adaic_Links) & " Links");
    for I in Class_Type loop
      case I is
        when Html_Text =>
          Ada.Text_Io.Put(Result_File, "%% HTML files              ");
        when Archived =>
          Ada.Text_Io.Put(Result_File, "%% Archived (ZIP, tar, gz) ");
        when Pdf =>
          Ada.Text_Io.Put(Result_File, "%% PDF files               ");
        when Ms_Word =>
          Ada.Text_Io.Put(Result_File, "%% Microsoft Word (.doc)   ");
        when Text =>
          Ada.Text_Io.Put(Result_File, "%% Text files              ");
        when Postscript =>
          Ada.Text_Io.Put(Result_File, "%% Postscript files        ");
        when Graphic =>
          Ada.Text_Io.Put(Result_File, "%% Graphic files           ");
        when Unclassified =>
          Ada.Text_Io.Put(Result_File, "%% Unclassified links      ");
      end case;
      Ada.Text_Io.Put_Line(Result_File, "%% Count:"
        & Fancy_Image(Class_Count(I), 7) & " Size:"
        & Fancy_Image(Class_Size(I), 12) & " or"
        & Fancy_Image(Class_K_Size(I), 9) & "K");
    end loop;
  end Print;

end Statistics;
